from dataclasses import dataclass
from typing import Literal

type Op1 = Literal['-']

type Op2 = Literal['-', '+']

type Expr = EInt | EVar | EOp1 | EOp2 | EInput

@dataclass
class EInt:
    val: int

@dataclass
class EVar:
    val: str

@dataclass
class EOp1:
    op: Op1
    e: Expr

@dataclass
class EOp2:
    e1: Expr
    op: Op2
    e2: Expr

@dataclass
class EInput:
    pass

type Stmt = SExpr | SAssign | SPrint

@dataclass
class SExpr:
    e: Expr

@dataclass
class SAssign:
    x: str
    e: Expr

@dataclass
class SPrint:
    e: Expr

type Prog = list[Stmt]

example: Prog = [
    SAssign('x', EInput()),
    SAssign('y', EOp2(EInput(), '+', EVar('x'))),
    SPrint(
        EOp2(
            EOp2(EVar('x'), '+', EVar('y')),
            '-',
            EInt(5)
        )
    )
]

@dataclass
class EvalError(Exception):
    s: str

def eval_prog(p: Prog):
    env: dict[str, int] = dict()
    for s in p:
        eval_stmt(env, s)

def eval_stmt(env: dict[str, int], s: Stmt):
    match s:
        case SExpr(e):
            eval_expr(env, e)
        case SAssign(x, e):
            env[x] = eval_expr(env, e)
        case SPrint(e):
            print(eval_expr(env, e))

def eval_expr(env: dict[str, int], e: Expr) -> int:
    match e:
        case EInt(i):
            return i
        case EVar(x):
            if x in env:
                return env[x]
            else:
                raise EvalError(f"Variable {x} is undefined.")
        case EOp1(op, e):
            v = eval_expr(env, e)
            match op:
                case '-':
                    return -v
        case EOp2(e1, op, e2):
            v1 = eval_expr(env, e1)
            v2 = eval_expr(env, e2)
            match op:
                case '+':
                    return v1 + v2
                case '-':
                    return v1 - v2
        case EInput():
            try:
                return int(input('> '))
            except ValueError:
                raise EvalError("Input was not an integer")

try:
    eval_prog(example)
except EvalError as err:
    print(f"Evaluation Error: {err.s}")
